from pallets_sphinx_themes import get_version

# Project --------------------------------------------------------------

project = "Pallets-Sphinx-Themes"
copyright = "2007 Pallets"
author = "Pallets"
release, version = get_version("pallets-sphinx-themes")

# General --------------------------------------------------------------

default_role = "code"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.log_cabinet",
    "myst_parser",
    "pallets_sphinx_themes",
]
autodoc_member_order = "bysource"
autodoc_typehints = "description"
autodoc_preserve_defaults = True
extlinks = {
    "issue": ("https://github.com/pallets/pallets-sphinx-themes/issues/%s", "#%s"),
    "pr": ("https://github.com/pallets/pallets-sphinx-themes/pull/%s", "#%s"),
}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}
myst_enable_extensions = [
    "fieldlist",
]
myst_heading_anchors = 2

# HTML -----------------------------------------------------------------

html_theme = "flask"
html_theme_options = {"index_sidebar_logo": False}
html_sidebars = {
    "index": ["project.html", "localtoc.html", "searchbox.html", "ethicalads.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html", "ethicalads.html"],
}
singlehtml_sidebars = {"index": ["project.html", "localtoc.html", "ethicalads.html"]}
html_title = f"{project} Documentation ({version})"
html_copy_source = False
html_show_sourcelink = False
html_show_copyright = False
html_use_index = False
html_domain_indices = False



import inspect
import os
import re
import sys
import textwrap
from collections import namedtuple
from importlib import metadata as importlib_metadata
from urllib.parse import urlsplit

from sphinx.application import Sphinx
from sphinx.builders.dirhtml import DirectoryHTMLBuilder
from sphinx.errors import ExtensionError

from .theme_check import only_pallets_theme
from .theme_check import set_is_pallets_theme


def setup(app):
    base = os.path.join(os.path.dirname(__file__), "themes")

    for name in os.listdir(base):
        path = os.path.join(base, name)

        if os.path.isdir(path):
            app.add_html_theme(name, path)

    app.add_config_value("is_pallets_theme", None, "html")
    app.add_config_value("rtd_canonical_path", "/en/stable/", "html")
    app.add_config_value("version_banner", True, "html")

    # Use the sphinx-notfound-page extension to generate a 404 page with valid
    # URLs. Only configure it if it's not already configured.
    if "notfound.extension" not in app.config.extensions:
        app.config.extensions.append("notfound.extension")
        app.config.notfound_context = {
            "title": "Page Not Found",
            "body": """<h1>Page Not Found</h1>
            <p>
              The page you requested does not exist. You may have followed a bad
              link, or the page may have been moved or removed.
            """,
        }

        if "READTHEDOCS" not in os.environ:
            # Disable the default prefix outside of Read the Docs.
            app.config.notfound_urls_prefix = None

    app.connect("builder-inited", set_is_pallets_theme)
    app.connect("builder-inited", find_base_canonical_url)
    app.connect("builder-inited", add_theme_files)
    app.connect("html-page-context", canonical_url)

    try:
        app.connect("autodoc-skip-member", skip_internal)
        app.connect("autodoc-process-docstring", cut_module_meta)
    except ExtensionError:
        pass

    from .themes import click as click_ext
    from .themes import jinja as jinja_ext

    click_ext.setup(app)
    jinja_ext.setup(app)

    own_release, _ = get_version(__name__)
    return {"version": own_release, "parallel_read_safe": True}


@only_pallets_theme()
def find_base_canonical_url(app: Sphinx) -> None:
    """When building on Read the Docs, build the base canonical URL from the
    environment variable if it's not given in the config. Replace the path with
    ``rtd_canonical_path``, which defaults to ``/en/stable/``.
    """
    if app.config.html_baseurl:
        return

    if "READTHEDOCS_CANONICAL_URL" in os.environ:
        parts = urlsplit(os.environ["READTHEDOCS_CANONICAL_URL"])
        path = app.config.rtd_canonical_path
        app.config.html_baseurl = f"{parts.scheme}://{parts.netloc}{path}"


@only_pallets_theme()
def add_theme_files(app: Sphinx) -> None:
    # Add the JavaScript for the version warning banner if ``version_banner`` is
    # enabled. On Read the Docs, don't include it for stable or PR builds.
    # Include the project and version as data attributes that the script will
    # access. The project name is assumed to be the PyPI name, and is normalized
    # to avoid a redirect.
    rtd_version = os.environ.get("READTHEDOCS_VERSION")
    rtd_version_type = os.environ.get("READTHEDOCS_VERSION_TYPE")

    if app.config.version_banner and (
        rtd_version is None  # not on read the docs
        or (rtd_version != "stable" and rtd_version_type in {"branch", "tag"})
    ):
        app.add_js_file(
            "describe_version.js",
            **{
                "data-project": re.sub(r"[-_.]+", "-", app.config.project).lower(),
                "data-version": app.config.version,
            },
        )


@only_pallets_theme()
def canonical_url(app: Sphinx, pagename, templatename, context, doctree):
    """Sphinx 1.8 builds a canonical URL if ``html_baseurl`` config is
    set. However, it builds a URL ending with ".html" when using the
    dirhtml builder, which is incorrect. Detect this and generate the
    correct URL for each page.
    """
    base = app.config.html_baseurl

    if (
        not base
        or not isinstance(app.builder, DirectoryHTMLBuilder)
        or not context["pageurl"]
        or not context["pageurl"].endswith(".html")
    ):
        return

    # Fix pageurl for dirhtml builder if this version of Sphinx still
    # generates .html URLs.
    target = app.builder.get_target_uri(pagename)
    context["pageurl"] = base + target


@only_pallets_theme()
def skip_internal(app, what, name, obj, skip, options):
    """Skip rendering autodoc when the docstring contains a line with
    only the string `:internal:`.
    """
    docstring = inspect.getdoc(obj) or ""

    if skip or re.search(r"^\s*:internal:\s*$", docstring, re.M) is not None:
        return True


@only_pallets_theme()
def cut_module_meta(app, what, name, obj, options, lines):
    """Don't render lines that start with ``:copyright:`` or
    ``:license:`` when rendering module autodoc. These lines are useful
    meta information in the source code, but are noisy in the docs.
    """
    if what != "module":
        return

    lines[:] = [
        line for line in lines if not line.startswith((":copyright:", ":license:"))
    ]


def get_version(name, version_length=2, placeholder="x"):
    """Ensures that the named package is installed and returns version
    strings to be used by Sphinx.

    Sphinx uses ``version`` to mean an abbreviated form of the full
    version string, which is called ``release``. In ``conf.py``::

        release, version = get_version("Flask")
        # release = 1.0.x, version = 1.0.3.dev0

    :param name: Name of package to get.
    :param version_length: How many values from ``release`` to use for
        ``version``.
    :param placeholder: Extra suffix to add to the version. The default
        produces versions like ``1.2.x``.
    :return: ``(release, version)`` tuple.
    """
    try:
        release = importlib_metadata.version(name)
    except ImportError:
        print(
            textwrap.fill(
                f"'{name}' must be installed to build the documentation."
                " Install from source using `pip install -e .` in a virtualenv."
            )
        )
        sys.exit(1)

    version = ".".join(release.split(".", version_length)[:version_length])

    if placeholder:
        version = f"{version}.{placeholder}"

    return release, version


#: ``(title, url)`` named tuple that will be rendered with
ProjectLink = namedtuple("ProjectLink", ("title", "url"))



import contextlib
import shlex
import subprocess
import sys
import tempfile
from functools import partial

import click
from click.testing import CliRunner
from click.testing import EchoingStdin
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from sphinx.domains import Domain


class EofEchoingStdin(EchoingStdin):
    """Like :class:`click.testing.EchoingStdin` but adds a visible
    ``^D`` in place of the EOT character (``\x04``).

    :meth:`ExampleRunner.invoke` adds ``\x04`` when
    ``terminate_input=True``.
    """

    def _echo(self, rv):
        eof = rv[-1] == b"\x04"[0]

        if eof:
            rv = rv[:-1]

        if not self._paused:
            self._output.write(rv)

            if eof:
                self._output.write(b"^D\n")

        return rv


@contextlib.contextmanager
def patch_modules():
    """Patch modules to work better with :meth:`ExampleRunner.invoke`.

    ``subprocess.call` output is redirected to ``click.echo`` so it
    shows up in the example output.
    """
    old_call = subprocess.call

    def dummy_call(*args, **kwargs):
        with tempfile.TemporaryFile("wb+") as f:
            kwargs["stdout"] = f
            kwargs["stderr"] = f
            rv = subprocess.Popen(*args, **kwargs).wait()
            f.seek(0)
            click.echo(f.read().decode("utf-8", "replace").rstrip())
        return rv

    subprocess.call = dummy_call

    try:
        yield
    finally:
        subprocess.call = old_call


class ExampleRunner(CliRunner):
    def __init__(self):
        super().__init__(echo_stdin=True)
        self.namespace = {"click": click, "__file__": "dummy.py"}

    @contextlib.contextmanager
    def isolation(self, *args, **kwargs):
        iso = super().isolation(*args, **kwargs)

        with iso as streams:
            try:
                buffer = sys.stdin.buffer
            except AttributeError:
                buffer = sys.stdin

            # FIXME: We need to replace EchoingStdin with our custom
            # class that outputs "^D". At this point we know sys.stdin
            # has been patched so it's safe to reassign the class.
            # Remove this once EchoingStdin is overridable.
            buffer.__class__ = EofEchoingStdin
            yield streams

    def invoke(
        self,
        cli,
        args=None,
        prog_name=None,
        input=None,
        terminate_input=False,
        env=None,
        _output_lines=None,
        **extra,
    ):
        """Like :meth:`CliRunner.invoke` but displays what the user
        would enter in the terminal for env vars, command args, and
        prompts.

        :param terminate_input: Whether to display "^D" after a list of
            input.
        :param _output_lines: A list used internally to collect lines to
            be displayed.
        """
        output_lines = _output_lines if _output_lines is not None else []

        if env:
            for key, value in sorted(env.items()):
                value = shlex.quote(value)
                output_lines.append(f"$ export {key}={value}")

        args = args or []

        if prog_name is None:
            prog_name = cli.name.replace("_", "-")

        output_lines.append(f"$ {prog_name} {shlex.join(args)}".rstrip())
        # remove "python" from command
        prog_name = prog_name.rsplit(" ", 1)[-1]

        if isinstance(input, (tuple, list)):
            input = "\n".join(input) + "\n"

            if terminate_input:
                input += "\x04"

        result = super().invoke(
            cli=cli, args=args, input=input, env=env, prog_name=prog_name, **extra
        )
        output_lines.extend(result.output.splitlines())
        return result

    def declare_example(self, source):
        """Execute the given code, adding it to the runner's namespace."""
        with patch_modules():
            code = compile(source, "<docs>", "exec")
            exec(code, self.namespace)

    def run_example(self, source):
        """Run commands by executing the given code, returning the lines
        of input and output. The code should be a series of the
        following functions:

        *   :meth:`invoke`: Invoke a command, adding env vars, input,
            and output to the output.
        *   ``println(text="")``: Add a line of text to the output.
        *   :meth:`isolated_filesystem`: A context manager that changes
            to a temporary directory while executing the block.
        """
        code = compile(source, "<docs>", "exec")
        buffer = []
        invoke = partial(self.invoke, _output_lines=buffer)

        def println(text=""):
            buffer.append(text)

        exec(
            code,
            self.namespace,
            {
                "invoke": invoke,
                "println": println,
                "isolated_filesystem": self.isolated_filesystem,
            },
        )
        return buffer

    def close(self):
        """Clean up the runner once the document has been read."""
        pass


def get_example_runner(document):
    """Get or create the :class:`ExampleRunner` instance associated with
    a document.
    """
    runner = getattr(document, "click_example_runner", None)
    if runner is None:
        runner = document.click_example_runner = ExampleRunner()
    return runner


class DeclareExampleDirective(Directive):
    """Add the source contained in the directive's content to the
    document's :class:`ExampleRunner`, to be run using
    :class:`RunExampleDirective`.

    See :meth:`ExampleRunner.declare_example`.
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False

    def run(self):
        doc = ViewList()
        runner = get_example_runner(self.state.document)

        try:
            runner.declare_example("\n".join(self.content))
        except BaseException:
            runner.close()
            raise

        doc.append(".. sourcecode:: python", "")
        doc.append("", "")

        for line in self.content:
            doc.append(" " + line, "")

        node = nodes.section()
        self.state.nested_parse(doc, self.content_offset, node)
        return node.children


class RunExampleDirective(Directive):
    """Run commands from :class:`DeclareExampleDirective` and display
    the input and output.

    See :meth:`ExampleRunner.run_example`.
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False

    def run(self):
        doc = ViewList()
        runner = get_example_runner(self.state.document)

        try:
            rv = runner.run_example("\n".join(self.content))
        except BaseException:
            runner.close()
            raise

        doc.append(".. sourcecode:: shell-session", "")
        doc.append("", "")

        for line in rv:
            doc.append(" " + line, "")

        node = nodes.section()
        self.state.nested_parse(doc, self.content_offset, node)
        return node.children


class ClickDomain(Domain):
    name = "click"
    label = "Click"
    directives = {"example": DeclareExampleDirective, "run": RunExampleDirective}

    def merge_domaindata(self, docnames, otherdata):
        # Needed to support parallel build.
        # Not using self.data -- nothing to merge.
        pass


def delete_example_runner_state(app, doctree):
    """Close and remove the :class:`ExampleRunner` instance once the
    document has been read.
    """
    runner = getattr(doctree, "click_example_runner", None)

    if runner is not None:
        runner.close()
        del doctree.click_example_runner


def setup(app):
    app.add_domain(ClickDomain)
    app.connect("doctree-read", delete_example_runner_state)
